# -*- Mode: Python; test-case-name: savon.test.test_actor -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Savon - sysadmin versioning
# Copyright (C) 2006 Thomas Vander Stichele <thomas at apestaart dot org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.

"""
The actor implements the action that savon can take.
"""

import errno
import os
import commands
import sys
import re
import glob

import pysvn

from savon import overlay, client, state
from savon.util import log, util

class Actor(log.Loggable):
    """
    @ivar overlays: dict of name -> overlay
    @type overlays: dict of str -> L{overlay.Overlay}
    """
    logCategory = 'actor'

    def __init__(self, quiet=False, statedir=None, stdout=sys.stdout,
        stderr=sys.stderr):
        """
        @param statedir: savon state directory
                         defaults to $HOME/.savon
        """
        self._client = client.Client()
        self._client.pysvn.callback_notify = self.svnCallbackNotify
        self.quiet = quiet
        self._state = state.State(statedir)
        self.statedir = self._state.dir
        self.debug('created Actor with state dir %r' % self.statedir)

        self.overlays = {} # name -> overlay
        self._pathToOverlay = {} # live path -> overlay
        self._updateOverlays()
        self.stdout = stdout
        self.stderr = stderr
        self._notifyData = [] # make sure we have an array for notifies even
                              # when we're not interested

    def svnCallbackNotify(self, kwargs):
        # surprisingly, the callback gets called right *before* the change
        # we get notified of.  So all we can do is queue and react later
        self.log('svnCallbackNotify: kwargs %r' % kwargs)
        self._notifyData.append(kwargs)

    def _setLivePathStatus(self, livePath, position, letter):
        if not livePath in self._notifyOutput.keys():
            self._notifyOutput[livePath] = [' ', ] * 6
        self._notifyOutput[livePath][position] = letter

    def handleNotify(self, kwargs, overlay, layer):
        # kwargs has kind, content_state, error, action, path, revision,
        a = kwargs['action']
        forwardPath = kwargs['path']
        self.log('handleNotify: forwardPath %s, kwargs %r' % (
            forwardPath, kwargs))
        vPath = layer.versionedPathFromForward(forwardPath)
        livePath = overlay.livePathFromVersioned(vPath)
        if a == pysvn.wc_notify_action.update_update:
            cs = kwargs['content_state']
            ps = kwargs['prop_state']
            # when a file gets updated, its parent directory gets updated
            # as well, even though it may not be in our dict.
            # FIXME: for now, just ignore it, because this parent dir
            # is not guaranteed to be toplevel
            if forwardPath not in self._forwardPathDict:
                child = False
                for key in self._forwardPathDict.keys():
                    if forwardPath == os.path.dirname(key):
                        child = True
                if child:
                    self.debug(
                        "forwardPath %r not in our expected paths, "
                        "but parent of one, ignoring" % forwardPath)
                else:
                    self.warning("Unknown forward path %s" % forwardPath)
            else:
                overlay, layer, reverse = self._forwardPathDict[forwardPath]
            # handle content state
            if cs == pysvn.wc_notify_state.unchanged: 
                pass
            elif cs == pysvn.wc_notify_state.changed: 
                # reverse this live path if it is
                self._setLivePathStatus(livePath, 0, 'U')
                self.debug('forwardPath %s changed' % forwardPath)
                if reverse:
                    self.debug('forwardPath %r is toplevel for %s, reverse' % (
                        forwardPath, livePath))
                    layer.reverse(livePath, vPath)
                else:
                    self.debug('forwardPath %r is not toplevel for %s' % (
                        forwardPath, livePath))
            elif cs == pysvn.wc_notify_state.conflicted: 
                self.debug('forwardPath %s conflicted' % forwardPath)
                self._setLivePathStatus(livePath, 0, 'C')
                self._state.appendOverlayConflicts(overlay.name, [livePath, ])
            elif cs == pysvn.wc_notify_state.merged: 
                self.debug('forwardPath %s state merged' % forwardPath)
                self._setLivePathStatus(livePath, 0, 'G')
                if reverse:
                    self.debug('forwardPath %r is toplevel for %s, reverse' % (
                        forwardPath, livePath))
                    layer.reverse(livePath, vPath)
            elif cs == pysvn.wc_notify_state.unknown: 
                self.debug('forwardPath %s state unknown' % forwardPath)
            else:
                self.error('Unhandled content state %r for kwargs %r' % (
                    cs, kwargs))
            # handle properties state
            if ps == pysvn.wc_notify_state.unchanged: 
                pass
            elif ps == pysvn.wc_notify_state.changed: 
                # reverse this live path if it is
                self._setLivePathStatus(livePath, 1, 'U')
                self.debug('forwardPath %s changed' % forwardPath)
                if reverse:
                    self.debug('forwardPath %r is toplevel for %s, reverse' % (
                        forwardPath, livePath))
                    layer.reverse(livePath, vPath)
                else:
                    self.debug('forwardPath %r is not toplevel for %s' % (
                        forwardPath, livePath))
            elif ps == pysvn.wc_notify_state.conflicted: 
                self.debug('forwardPath %s conflicted' % forwardPath)
                self._setLivePathStatus(livePath, 1, 'C')
                self._state.appendOverlayConflicts(overlay.name, [livePath, ])
            elif ps == pysvn.wc_notify_state.merged: 
                self.debug('forwardPath %s state merged' % forwardPath)
                self._setLivePathStatus(livePath, 1, 'G')
                if reverse:
                    self.debug('forwardPath %r is toplevel for %s, reverse' % (
                        forwardPath, livePath))
                    layer.reverse(livePath, vPath)
            elif ps == pysvn.wc_notify_state.unknown: 
                self.debug('forwardPath %s props state unknown' % forwardPath)
            else:
                self.error('Unhandled props state %r for kwargs %r' % (
                    ps, kwargs))
        elif a == pysvn.wc_notify_action.restore:
            # reverse this live path if it is
            self.debug('forwardPath %r restored, reverse to %s' % (
                        forwardPath, livePath))
            layer.reverse(livePath, vPath)
        elif a == pysvn.wc_notify_action.update_add:
            # a new path is not in our dict, and can result in conflicts
            self.debug('new path added: %s' % forwardPath)
            ret = self.checkNewLivePathConflict(overlay, layer, livePath,
                    force=False)
            if ret:
                self.debug('newly added path %s is conflicting' % forwardPath)
                self._setLivePathStatus(livePath, 0, 'C')
                self._state.appendOverlayConflicts(overlay.name, [livePath, ])
            else:
                self._setLivePathStatus(livePath, 0, 'A')
        elif a == pysvn.wc_notify_action.update_delete:
            # it may not actually be deleted from our forward Path, since we
            # may have synced a newer live copy.  Delete it here to be sure.
            try:
                os.unlink(forwardPath)
            except OSError:
                pass
            # if the livePath is in a higher layer, nothing is needed
            # if it is in a lower layer, we reverse the lower layer version
            # if it's nowhere else, we delete the live copy
            # a new path is not in our dict, and can result in conflicts
            topLayer = overlay.exists(livePath)
            if not topLayer:
                self.debug('%s not in overlay anymore' % livePath)
                self._setLivePathStatus(livePath, 0, 'D')
                if os.path.isdir(livePath):
                    # FIXME: catch Exception for when the dir is not empty
                    os.rmdir(livePath)
                else:
                    os.unlink(livePath)
            else:
                if overlay.order(layer, topLayer) > 0:
                    # current layer is above new top layer, so reverse
                    self.debug('%s now comes from lower layer %r, reversing' % (
                        livePath, topLayer))
                    self._setLivePathStatus(livePath, 0, 'U')
                    topLayer.reverse(livePath, vPath)
                else:
                    # current layer is below new top layer, so no change
                    self.debug('%s is in higher layer %r, ignoring' % (
                        livePath, topLayer))
                    self._setLivePathStatus(livePath, 0, '^')
        elif a == pysvn.wc_notify_action.skip:
            # Skip happens when we asked to update a subpath of a non-versioned
            # path (but not when we ask to update a non-versioned path)
            self.debug('Skipped')
        elif a == pysvn.wc_notify_action.update_completed:
            self.debug('Update completed')
        else:
            self.error('Unhandled notify action %r for kwargs %r' % (a, kwargs))

    def create(self, URL, name, livePath='/'):
        """
        Create a new overlay in the given repository, with the given name,
        for the given live path.

        @param URL:      URL of Subversion repository to create the overlay in
        @type  URL:      str
        @param name:     name of the overlay to create
        @type  name:     str
        @param livePath: local path to create the overlay for
        @type  livePath: str
        """
        self.debug('create(%s, %s, %s)' % (URL, name, livePath))

        ### first test prerequisites are met

        # livePath should not end with a separator, unless it is the root
        assert livePath is os.path.sep or not livePath.endswith(os.path.sep)
        self.debug('Overlay for path %s' % livePath)

        # verify that the base URL exists
        if not self._client.existsURL(URL):
            self.error('URL %s does not exist' % URL)

        overlaysURL = '%s/overlays' % URL
        if not self._client.existsURL(overlaysURL):
            self._client.mkdir(overlaysURL,
                "savon: creating overlays directory")

        # verify that the overlay URL does not exist yet
        overlayURL = "%s/overlays/%s" % (URL, name)
        if self._client.existsURL(overlayURL):
            self.error('Overlay %s already exists at %s' % (name, URL))

        # verify that the local state does not exist
        if self._state.existsOverlay(name):
            self.error('Overlay %s already exists' % name)

        ### now proceed

        msg = "savon: creating overlay for %s" % name
        # create the repository URL
        self._client.mkdir(overlayURL, msg)

        # do a checkout
        overlayDir = self._state.getOverlayDir(name)
        self.system("svn co %s %s" % (overlayURL, overlayDir))

        # write the state files
        self._state.writeOverlayPath(name, livePath)
        self._state.writeOverlayLayers(name, [])
        self._state.writeOverlayCurrent(name, name)

        # update our internal state so we can get the overlay in addLayer
        # FIXME: a little ugly to do this twice
        self._updateOverlays() # live path -> overlay

        # add a first layer
        self.addLayer(URL, name, name)

        # update our internal state
        self._updateOverlays() # live path -> overlay

        # commit state files
        self.debug('committing overlay state')
        self.system("svn add %s/path" % overlayDir)
        self.system("svn add %s/layers" % overlayDir)
        self.system("svn commit -m '%s' %s" % (msg, overlayDir))

    def addLayer(self, URL, overlayName, layerName, quiet=False):
        """
        Create a new layer for the given overlay in the given repository,
        with the given name.

        @param URL:         URL of the Subversion repository to create the
                            layer in
        @type  URL:         str
        @param overlayName: name of the overlay for which to create a layer.
        @param layerName:   name of the layer to create
        @type  layerName:   str
        @param quiet:       whether to suppress output
        """
        self.debug('addLayer(%s, %s, %s)' % (URL, overlayName, layerName))

        ### first test prerequisites are met

        # verify that the base URL exists
        if not self._client.existsURL(URL):
            self.error('URL %s does not exist' % URL)

        layersURL = '%s/layers' % URL
        if not self._client.existsURL(layersURL):
            self._client.mkdir(layersURL,
                "savon: creating layers directory")
            if not quiet:
                self.stdout.write("Created layers URL '%s'\n" % layersURL)

        # verify that the local state does not exist
        if self._state.existsLayer(layerName):
            self.error('Local layer %s already exists' % layerName)

        ### now proceed

        msg = "savon: creating layer for %s" % layerName
        layerURL = "%s/layers/%s" % (URL, layerName)
        if self._client.existsURL(layerURL):
            self.debug('Layer %s already exists at %s' % (layerName, URL))
        else:
            # create the repository URL
            self._client.mkdir(layerURL, msg)
            if not quiet:
                self.stdout.write("Created layer URL '%s'\n" % layerURL)

        # do a checkout
        localLayerDir = os.path.join(self.statedir, 'layers', layerName)
        self.system("svn co %s %s" % (layerURL,
            os.path.join(localLayerDir, 'forward')))

        # add the new layer to the overlay's state
        # FIXME: move this code into the Overlay object ?
        # note that we will get an incomplete object at this point
        o = self.getOverlayByName(overlayName)
        layers = []

        if not o.current:
            # we are above the top, so put it first
            layers.append(layerName)

        for l in o.layers:
            layers.append(l.name)
            # if we're at the current layer, insert ours below
            if o.current and l == o.current:
                layers.append(layerName)

        self._state.writeOverlayLayers(overlayName, layers)
        self._state.writeOverlayCurrent(overlayName, layerName)

        # update our internal state
        self.debug('calling _updateOverlays')
        self._updateOverlays() # live path -> overlay

        # forward the "root" node of this overlay to this layer
        o = self.getOverlayByName(overlayName)
        forwardPath = o.forward(o.path, layer=o.current)
        self._client.updateSavonProperties(forwardPath)

        overlayDir = self._state.getOverlayDir(o.name)
        if not quiet:
            self.stdout.write("Sending overlay '%s'\n" % o.name)
        self.system("svn commit -m '%s' %s" % (msg, overlayDir))
        revision = self._client.getRevision(overlayDir)
        if not quiet:
            self.stdout.write("Commited revision %d.\n" % revision)

        layerDir = self._state.getLayerForwardDir(o.current.name)
        if not quiet:
            self.stdout.write("Sending layer '%s'\n" % o.current.name)
        self.system("svn commit -m '%s' %s" % (msg, layerDir))
        revision = self._client.getRevision(layerDir)
        if not quiet:
            self.stdout.write("Commited revision %d.\n" % revision)

    def system(self, command, quiet=None):
        """
        Execute the given command.
        Respect quietness.
        Error out if the return value is not 0.
        """
        self.debug('Executing command %r' % command)
        if quiet is None:
            quiet = self.quiet
        if quiet:
            command += "> /dev/null 2>&1"
        value = os.system(command)
        if value != 0:
            self.error('Could not execute %s' % command)

    def getOverlaysDictForLivePaths(self, livePaths):
        """
        For the given livePaths, get a dict of overlays they're managed by.

        @returns: dict of overlay -> list of livePath
        @rtype:   dict of L{savon.overlay.Overlay} -> list of str
        """
        # I hate bugs where I pass a str as a list argument
        assert not isinstance(livePaths, str), \
            "getOverlaysDictForLivePaths takes a list, not str"
        self.debug('getOverlaysDictForLivePaths(%r)' % livePaths)
        overlays = {} # dict of overlay -> list of live paths
        for livePath in livePaths:
            o = self.getOverlayByPath(livePath)
            self.debug('getOverlaysDictForLivePaths: '
                'found overlay %r for livePath %r' % (o, livePath))
            if o not in overlays.keys():
                overlays[o] = []
            overlays[o].append(livePath)
        self.debug('overlays: %r' % overlays)
        return overlays
 
    def forwardLayersForOverlaysDict(self, overlaysDict, recurse=True):
        """
        For the given overlays dict as returned by
        L{getOverlaysDictForLivePaths}, get the layers they (and possibly all
        their children) are in, and forward all versioned paths.

        @returns: dict of layer -> list of forwardPath
        @rtype:   dict of L{savon.layer.Layer} -> list of str
        """
        return self._updateLayersForOverlaysDict(
            overlaysDict, recurse, forward=True)

    def reverseLayersForOverlaysDict(self, overlaysDict, recurse=True):
        """
        For the given overlays dict as returned by
        L{getOverlaysDictForLivePaths}, get the layers they (and possibly all
        their children) are in, and and reverse all versioned paths
        (ie. copy from the repository version).

        @returns: dict of layer -> list of livePath
        @rtype:   dict of L{savon.layer.Layer} -> list of str
        """
        return self._updateLayersForOverlaysDict(
            overlaysDict, recurse, forward=False)

    def _updateLayersForOverlaysDict(self, overlaysDict, recurse=True,
        forward=True):
        # forward True:  live -> repo
        # forward False: repo -> live
        layers = {} # dict of layer -> list of live paths

        # create mapping of layer -> live paths, and sync while we're at it
        for o, livePaths in overlaysDict.items():
            for livePath in livePaths:
                self.debug('_updateLayersForOverlaysDict: '
                    'finding versioned live paths in livePath %s' %
                    livePath)
                versionedLivePaths = o.findVersionedLivePaths(livePath)
                for l, lPaths in versionedLivePaths.items():
                    self.debug('layer %r has live paths %r' % (l, lPaths))
                    # update our mapping
                    if l not in layers.keys():
                        layers[l] = []
                    # sync the state, and replace paths with forward paths
                    for p in lPaths:
                        f = None
                        if forward:
                            # if a path does not exist, we emulate svn's
                            # behaviour - restore from repository
                            if not os.path.lexists(p):
                                f = o.reverse(p, layer=l)
                                self.stdout.write("Restored '%s'\n" % p)
                                self.stdout.flush()
                            else:
                                f = o.forward(p, layer=l)
                                # if the forward path was in svn conflict,
                                # it is now resolved
                                status_list = self._client.status(
                                    f, recurse=False)
                                for s in status_list:
                                    # for dirs, we get all their entries too
                                    if s.path != f:
                                        continue
                                    if s.text_status == pysvn.wc_status_kind.conflicted:
                                        self._client.resolved(f, recurse=False)
                                        self.debug('svn resolved %s' % f)
                            o.updateSavonProperties(p, layer=l)
                        else:
                            o.updateFromSavonProperties(p, layer=l)
                            f = o.reverse(p, layer=l)
                        if recurse or p == livePath: 
                            self.log('Mapping forward path %r to layer %r' % (
                                f, l))
                            layers[l].append(f)
        self.debug('layers: %r' % layers)
        return layers

    def getConflictsForLivePaths(self, livePaths):
        """
        Get a list of files that are in conflict given the livePaths.

        @rtype: list of str
        """
        d = self.getOverlaysDictForLivePaths(livePaths)
        keep = []
        for o, livePaths in d.items():
            if not self._state.existsOverlayFile(o.name, 'conflicts'):
                continue
            conflicts = self._state.readOverlayConflicts(o.name)
            self.debug('%d conflicts for overlay %r' % (len(conflicts), o))
            for livePath in livePaths:
                versionedLivePaths = o.findVersionedLivePaths(livePath)
                for l, paths in versionedLivePaths.items():
                    for p in paths:
                        if p in conflicts:
                            keep.append(p)
        return keep

        
    def add(self, livePaths):
        """
        Add the given live paths to version control in the current layer of the
        correct overlay.
        Does not work recursively, too much potential for mistakes.

        @param livePaths: list of absolute live paths
        @type  livePaths: list of str

        @return: a list of output lines to show as a result
        @rtype:  list of str
        """
        overlays = {}
        output = []

        self.debug('add(%r)' % livePaths)
        # verify all the files exist in the live area
        for livePath in livePaths:
            # a broken symlink still exists for savon's purposes
            if not os.path.lexists(livePath):
                self.error("Path '%s' does not exist." % livePath)

        for livePath in livePaths:
            overlays[livePath] = self.getOverlayByPath(livePath)

        # verify none of the files already exist in their respective overlay
        # in the currently active layer
        for livePath, o in overlays.items():
            self.debug('Checking path %s' % livePath)
            layer = o.exists(livePath)
            if layer and layer == o.current:
                self.error(
                    "Path '%s' is already in layer '%s' in overlay '%s'" % (
                        livePath, layer.name, o.name))

        # now we can add them to the current layer for each overlay
        # note that we need to order based on the order given, since we
        # typically add a parent before its child
        for livePath in livePaths:
            o = overlays[livePath]
            layer = o.current
            self.debug('Adding path %s to overlay %s, layer %s' % (
                livePath, o.name, layer))
            o.add(livePath)
            output.append("A         %s" % livePath)

        self.debug('add(%r): done, output %r' % (livePaths, output))
        return output

    def information(self, livePath):
        """
        Give information about the overlay.
        """
        overlaysDict = self.getOverlaysDictForLivePaths([livePath, ])
        o = overlaysDict.keys()[0]
        url = o.getURL()
        output = [
            "Overlay: %s" % o.name,
            "Overlay Path: %s" % o.path,
            "Repository Root: %s" % url,
            ""
        ]
        return output

    def status(self, livePaths, recurse=True, quiet=False):
        """
        Give a status on the given list of paths.
        This also takes into account unversioned paths, which will be shown
        with ?, if quiet is False.

        @param recurse: whether to recurse into subdirectories or not
        @param quiet:   whether to show as little as status as possible;
                        don't show unversioned paths for example

        @returns: ordered list of
                    (livePath, layerName, versioned, status, revision,
                    commit_author, commit_revision)
        @rtype:   list of (str, bool, str, int, str, int)
        """
        self.debug('status(%r)' % livePaths)
        ret = []

        # verify all the live paths exist in the live area
        for livePath in livePaths:
            if not os.path.lexists(livePath):
                self.error("Path '%s' does not exist." % livePath)

        overlaysDict = self.getOverlaysDictForLivePaths(livePaths)

        # loop per overlay
        for o, livePaths in overlaysDict.items():
            self.debug('overlay %r' % o)
            layers = self.forwardLayersForOverlaysDict({o: livePaths},
                recurse=recurse)

            # now find all livePaths that we should give status on
            allLivePaths = []
            for livePath in livePaths:
                result = [livePath]

                def walker(result, dirname, fnames):
                    self.debug('walker: dirname %s', dirname)
                    
                    # if dirname is not versioned, then don't do anything
                    layer = o.exists(dirname)
                    if not layer:
                        self.debug('walker: not versioned, returning')
                        # note that we need slice assignment to really
                        # remove all children
                        fnames[:] = []
                        return

                    for fname in fnames:
                        liveChild = os.path.join(dirname, fname)
                        result.append(liveChild)

                        # remove this child from the list if it is not versioned
                        # so we don't recurse into it
                        if os.path.isdir(liveChild):
                            layer = o.exists(liveChild)
                            if not layer:
                                self.debug('not recursing into %s', liveChild)
                                fnames.remove(fname)

                if recurse:
                    vPath = o.versionedPathFromLive(livePath)
                    layer = o.exists(livePath)
                    # FIXME: I think this call is just here to make sure we
                    # forward the path again to get an updated version
                    #if layer:
                    #    forwardPath = layer.forwardPathFromVersioned(vPath)
                    os.path.walk(livePath, walker, result)

                allLivePaths.extend(result)

            for livePath in allLivePaths:
                first = second = ' '
                layer = o.exists(livePath)
                if layer:
                    first = ' '
                    vPath = o.versionedPathFromLive(livePath)
                    forwardPath = layer.forwardPathFromVersioned(vPath)
                    self.log('svn status on %s' % forwardPath)
                    res = self._client.status(forwardPath)
                    for status in res:
                        # we only want the forwardPath we're dealing with;
                        # asking for status on a dir also gives status objects
                        # on its children
                        if status.path != forwardPath:
                            continue

                        # first look at text_status
                        if status.text_status == pysvn.wc_status_kind.normal:
                            pass
                        elif status.text_status == pysvn.wc_status_kind.added:
                            first = 'A'
                        elif status.text_status == pysvn.wc_status_kind.modified:
                            first = 'M'
                        else:
                            self.error(
                                "Unhandled text status for %s in layer %s: %r" % (
                                livePath, layer.name, status.text_status))

                        # second, look at prop_status
                        if status.prop_status == pysvn.wc_status_kind.none:
                            pass
                        elif status.prop_status == pysvn.wc_status_kind.normal:
                            pass
                        elif status.prop_status == pysvn.wc_status_kind.modified:
                            second = 'M'
                        elif status.prop_status == pysvn.wc_status_kind.conflicted:
                            second = 'C'
                        else:
                            self.error(
                                "Unhandled prop status for %s in layer %s: %r" % (
                                livePath, layer.name, status.prop_status))
                    state = '%s%s     ' % (first, second)
                    ret.append((livePath, layer.name, True, state,
                        status.entry.revision.number,
                        status.entry.commit_author,
                        status.entry.commit_revision.number))
                else:
                    # not versioned
                    if not quiet:
                        ret.append((livePath, '', False, '?      ',
                            -1, None, -1))

        ret.sort()

        self.debug('status: returning %d lines' % len(ret))
        return ret

    def checkNewLivePathConflict(self, overlay, layer, livePath,
        force=False):
        """
        Check if the newly checked out forward path for the given live
        path is conflicting or not.

        @returns: forward path if conflicting, or None if not
        @rtype: str or None
        """
        self.log('checkNewLivePathConflict(%r, %r, %s, %r)' % (
            overlay, layer, livePath, force))
        vPath = overlay.versionedPathFromLive(livePath)
        forwardPath = layer.forwardPathFromVersioned(vPath)
        # first update forward path from their savon properties
        self._client.updateFromSavonProperties(forwardPath)
        self.log('comparing livePath %s and forwardPath %s' % (
            livePath, forwardPath))

        # a non-existent livePath means we can safely make it live
        if not os.path.exists(livePath):
            self.debug('livePath %s does not exist, using repo version' %
                livePath)
            self.reverseLayersForOverlaysDict({overlay: [livePath, ]},
                recurse=False)
            return

        if not util.pathEquals(livePath, forwardPath):
            if not force:
                self.debug('livePath %s is different, conflicting' %
                    livePath)
                return livePath
            else:
                self.debug('livePath %s is different, using repo version' %
                    livePath)
                self.reverseLayersForOverlaysDict({overlay: [livePath, ]},
                    recurse=False)
        else:
            self.log('paths equal')

    def _checkoutLayer(self, URL, layerName):
        layerURL = "%s/layers/%s" % (URL, layerName)
        self.stdout.write('Checking out layer from %s ...\n' % layerURL)
        self.system("svn co %s %s" % (layerURL,
            self._state.getLayerForwardDir(layerName)))

    def checkout(self, URL, name, force=False, livePath=None):
        """
        Check out the given overlay from the given URL.

        @param URL:      URL of the savon repository
        @param force:    whether to resolve all conflicts by preferring the
                         repository version
        @param livePath: override the overlay's default live path.
        """
        self.debug('checkout(%s, %s, force=%r, livePath=%r)' % (
            URL, name, force, livePath))
        overlaysURL = '%s/overlays' % URL
        if not self._client.existsURL(overlaysURL):
            self.error("No overlays directory found at %s" % URL)

        # verify that the overlay URL exists
        overlayURL = "%s/overlays/%s" % (URL, name)
        if not self._client.existsURL(overlayURL):
            self.error("%s not found" % overlayURL)

        # verify that the local state does not exist
        if self._state.existsOverlay(name):
            self.error('Local overlay %s already exists' % name)

        # do a checkout of the overlays directory
        self.debug('Checking out overlayURL %s' % overlayURL)
        self.stdout.write('Checking out overlay from %s ...\n' % overlayURL)
        self.system("svn co %s %s" % (overlayURL,
            self._state.getOverlayDir(name)))

        # already create the conflicts file to make sure other commands
        # can't run
        self._state.writeOverlayConflicts(name, None)

        # handle live path override
        if livePath:
            self.debug("Overriding live path to %s" % livePath)
            self._state.writeOverrideOverlayPath(name, livePath)

        # now check out all the layers
        layerNames = self._state.readOverlayLayers(name)
        for layerName in layerNames:
            self._checkoutLayer(URL, layerName)

        # and make the top one current
        self._state.writeOverlayCurrent(name, layerNames[0])

        # update our internal state
        self._updateOverlays()

        o = self.overlays[name]
        versionedLivePaths = o.findVersionedLivePaths(o.path)

        conflicts = []

        # for every versioned file, check for differences
        # between the top layer's version and the live version
        for layer in o.layers:
            livePaths = versionedLivePaths[layer]
            for livePath in livePaths:
                ret = self.checkNewLivePathConflict(o, layer, livePath,
                    force=force)
                if ret:
                    conflicts.append(ret)

        if conflicts:
            self._state.writeOverlayConflicts(name, conflicts)
        else:
            self._state.deleteOverlayConflicts(name)

        self.debug('checkout(%s, %s, force=%r): %d conflicts' % (
            URL, name, force, len(conflicts)))
        return conflicts

    def commit(self, livePaths, message=None, layer=None):
        """
        Commit everything in the given paths to version control.
        will also store ownership/permissions as properties.

        @param message: The commit message.  If none given, an editor
                        will be opened for each layer commited.
        @type  message: str
        @param layer:   name of the layer to commit, or None for all.
        @type  layer:   str

        @rtype: L{pysvn.Revision} 
        """
        assert not isinstance(livePaths, str), \
            "Internal error: %r takes a list of str, not str" % self.commit
        if layer and not self._state.existsLayer(layer):
            self.error('Layer %s does not exist.' % layer)
            return

        self.debug('commit(%r, %r)' % (livePaths, message))
        overlaysDict = self.getOverlaysDictForLivePaths(livePaths)
        layers = self.forwardLayersForOverlaysDict(overlaysDict)

        # if a layer is specified, only deal with that one layer
        if layer:
            found = False
            for l, forwardPaths in layers.items():
                if l.name == layer:
                    found = True
                    layers = {l: forwardPaths}
                    break

            if not found:
                # no changes in the layer
                self.stdout.write("No changes in layer '%s'.\n" % layer)
                return

        # now layers is filled in and we can start committing
        for l, forwardPaths in layers.items():
            # make sure we commit in the right order
            if not forwardPaths:
                self.debug('No forwardPaths for layer %r' % l)
                continue
            # make sure that we have the parent for each forwardPath as well
            forwardDir = os.path.join(l.dir, 'forward')
            for p in forwardPaths[:]:
                while p != forwardDir:
                    p = os.path.dirname(p)
                    if p not in forwardPaths:
                        self.debug('Adding parent %s to forwardPaths' % p)
                        forwardPaths.append(p)
            forwardPaths.sort()

            if not self.quiet:
                self.stdout.write("Committing layer %s\n" % l.name)

            # now commit; we system svn so we get nice output
            # non-recursive so we don't descend into parents
            cmd = 'svn commit -N'
            if message:
                cmd += ' -m "%s"' % message
            cmd += ' ' + ' '.join(forwardPaths)
            self.debug('Running command %s' % cmd)
            self.system(cmd)
        return
        
        # FIXME: further todo
        #revision = self._client.checkin(paths, message)
        # theoretically after a commit an update should give you the
        # revision, for any of the paths
        # revision = self._client.update(paths[0])
        # return revision
        
        # build a list of commited paths
        # for path in paths:
        #     msgs = self._client.log(path,
        #        revision_start=revision,
        #        revision_end=revision,
        #        discover_changed_paths=True)

        # return revision

    def diff(self, livePaths, recurse=True):
        # FIXME: we should not print ourselves, but let our invoking Command
        # print
        self.debug('diff(%r)' % livePaths)
        overlaysDict = self.getOverlaysDictForLivePaths(livePaths)
        ret = []
        found = False

        # loop per overlay
        for o, livePaths in overlaysDict.items():
            self.debug('overlay %r' % o)
            layers = self.forwardLayersForOverlaysDict({o: livePaths},
                recurse=recurse)

            self.debug('overlay %r, layers %r' % (o, layers))
            # now layers is filled in and we can start diffing
            for l, forwardPaths in layers.items():
                outputs = []
                if forwardPaths:
                    found = True
                for forwardPath in forwardPaths:
                    command = 'svn diff --non-recursive %s' % forwardPath
                    self.debug('Executing %r' % command)
                    (status, output) = commands.getstatusoutput(command)

                    # a non-recursive diff on a directory still gives diffs
                    # of files in that directory.  We need to remove those
                    # since at best they would be listed twice, but at worst
                    # the file could be in a different layer and thus the diff
                    # of that file in this layer would not be relevant.
                    if os.path.isdir(forwardPath):
                        # if the directory itself does not have diffs,
                        # the output will start with the first child; so
                        # we need to search for an Index: line or
                        # Property changes on: line for a child

                        firstChildIndexOffset = output.find("Index: %s" %
                            os.path.join(forwardPath, ''))
                        if firstChildIndexOffset > -1:
                            offset = firstChildIndexOffset
                            self.debug(
                                'Found child Index for forwardPath %s at %d' % (
                                    forwardPath, firstChildIndexOffset))

                        firstChildPropertyOffset = output.find(
                            "\nProperty changes on: %s" %
                            os.path.join(forwardPath, ''))
                        if firstChildPropertyOffset > -1:
                            self.debug(
                                'Found child Property at %d' %
                                    firstChildPropertyOffset)

                        offset = -1
                        if firstChildIndexOffset > -1:
                            offset = firstChildIndexOffset
                        if firstChildPropertyOffset > -1:
                            if firstChildPropertyOffset < offset \
                            or offset == -1:
                                offset = firstChildPropertyOffset

                        if offset > -1:
                            output = output[:offset]

                    # scrub the output of paths that reference the forward layer
                    if output:
                        self.debug('forwardPath %s has diffs' % forwardPath)
                        outputs.append(self._scrubDiffOutput(output,
                            l.forwardPathFromVersioned(os.path.sep),
                            o.path))
                if outputs:
                    ret.append("Diff of overlay %s, layer %s:" % (
                        o.name, l.name))
                    for output in outputs:
                        ret.append(output)

        if not found:
            ret.append("Path(s) %s not found in any layer." %
                " ".join(livePaths))

        return ret
                
    def _scrubDiffOutput(self, output, layerForwardPath, overlayLivePath):
        """
        Scrub paths in the output and replace leading layerForwardPath with
        leading overlayLivePath.
        """
        # FIXME: naive way for now
        # but ideally we would check for known strings, like Index:, ---, ...
        pattern = re.compile(layerForwardPath, re.MULTILINE)
        sub = overlayLivePath
        if overlayLivePath == os.path.sep:
            sub = ''
        return re.sub(pattern, sub, output)

    def resolveLive(self, livePaths):
        self.debug('resolveLive(%r)' % livePaths)
        overlaysDict = self.getOverlaysDictForLivePaths(livePaths)

        # loop per overlay
        for o, livePaths in overlaysDict.items():
            layers = self.forwardLayersForOverlaysDict({o: livePaths},
                recurse=False)

            self.debug('overlay %r, layers %r' % (o, layers))
            self._updateConflictsForLayersDict(o, layers)

    def resolveRepository(self, livePaths):
        """
        Resolve the given list of live paths by copying the repository version
        to the live version.

        @rtype:   list of str
        @returns: lines of output
        """
        # FIXME: allow recursive resolve
        res = []
        self.debug('resolveRepository(%r)' % livePaths)
        overlaysDict = self.getOverlaysDictForLivePaths(livePaths)

        # first svn revert each of the livePaths, to make sure we get
        # the repository version
        res = self.revert(livePaths)

        # loop per overlay
        for o, livePaths in overlaysDict.items():
            layers = self.reverseLayersForOverlaysDict({o: livePaths},
                recurse=False)

            self.debug('overlay %r, layers %r' % (o, layers))
            self._updateConflictsForLayersDict(o, layers)

        return res

    def resolveDiff(self, livePaths):
        """
        Returns the diffs that need resolving.
        """
        ret = []

        self.debug('resolveDiff(%r)' % livePaths)
        overlaysDict = self.getOverlaysDictForLivePaths(livePaths)
        for o, livePaths in overlaysDict.items():
            conflicts = self._state.readOverlayConflicts(o.name)
            for livePath in conflicts:
                show = False
                # only show what we asked for
                for l in livePaths:
                    if livePath.startswith(l):
                        show = True
                if not show:
                    continue

                layer = o.exists(livePath)
                vPath = o.versionedPathFromLive(livePath)
                forwardPath = layer.forwardPathFromVersioned(vPath)
                self.debug('showing diff between %s and %s' % (
                    livePath, forwardPath))
                # file differences
                cmd = "diff -au %s %s" % (forwardPath, livePath)
                self.debug('cmd %s' % cmd)
                output = commands.getoutput(cmd)
                ret.append(output + '\n')

                # property differences
                diff = util.diffPathProperties(forwardPath, livePath)
                if diff:
                    ret.append('\nProperty changes on: %s\n' % livePath)
                    ret.append('-' * 70 + '\n')
                    for key, (value1, value2) in diff.items():
                        ret.append('\nName: savon:%s\n' % key)
                        ret.append('    - %s\n' % value1)
                        ret.append('    + %s\n' % value2)

        return ret

    def resolveEdit(self, command, livePaths):
        self.debug('resolveEdit(%r)' % livePaths)
        overlaysDict = self.getOverlaysDictForLivePaths(livePaths)
        layers = {}
        for o, livePaths in overlaysDict.items():
            conflicts = self._state.readOverlayConflicts(o.name)
            conflicts = [c for c in conflicts if c in livePaths]
            for livePath in conflicts:
                layer = o.exists(livePath)
                if not layer in layers.keys():
                    layers[layer] = []
                vPath = o.versionedPathFromLive(livePath)
                forwardPath = layer.forwardPathFromVersioned(vPath)
                layers[layer].append(forwardPath)

                self.debug('editing diff between %s and %s' % (
                    livePath, forwardPath))
                cmd = "%s %s %s" % (command, livePath, forwardPath)
                self.debug('resolve edit using %s' % cmd)
                status = os.system(cmd)
                ret = os.WEXITSTATUS(status)
                if ret == 127:
                    self.error("%s is not installed." % command)
                if ret != 0:
                    self.error("Executing '%s' returned %d." % (cmd, ret))

            self._updateConflictsForLayersDict(o, layers)

    def revert(self, livePaths, recurse=False):
        """
        Revert local changes in the given paths.

        @type livePaths: list of str

        @rtype:   list of str
        @returns: lines of output
        """
        self.debug('revert(%r, recurse=%r)' % (livePaths, recurse))

        res = []
        overlaysDict = self.getOverlaysDictForLivePaths(livePaths)
        forwardPaths = {} # forwardPath -> overlay, livePath
        revertedPaths = [] # allows us to order results by incoming livePaths

        # to realize that there is a change to revert between live and
        # repository version, we first need to forward the livePath.

        # do the reverts
        for o, oLivePaths in overlaysDict.items():
            for livePath in oLivePaths:
                topLayerToLivePaths = o.findVersionedLivePaths(
                    livePath, onlyTop=True)
                for layer, vLivePaths in topLayerToLivePaths.items():
                    for vLivePath in vLivePaths:
                        o.forward(vLivePath)
                        vPath = o.versionedPathFromLive(vLivePath)
                        forwardPath = layer.forwardPathFromVersioned(vPath)
                        # store to handle the notifies
                        forwardPaths[forwardPath] = (o, vLivePath)
                        self.debug('revert: reverting forwardPath %s' %
                            forwardPath)
                        self._client.revert(forwardPath)

        # handle the notifies
        for kwargs in self._notifyData:
            if kwargs['action'] == pysvn.wc_notify_action.revert:
                forwardPath = kwargs['path']
                overlay, livePath = forwardPaths[forwardPath]
                self.debug('revert: reversing %s' % livePath)
                overlay.reverse(livePath)
                revertedPaths.append(livePath)

        # now output results in the order of the given livePaths
        for livePath in livePaths:
                if livePath in revertedPaths:
                    line = 'Reverted %s' % livePath
                    self.debug(line)
                    res.append(line)
            
        self._notifyData = []

        return res
 
    def _updateConflictsForLayersDict(self, o, layers):
        """
        @param layers: dict of layer -> list of forwardPath
        @type  layers: dict of L{savon.layer.Layer} -> list of str
        """
        # now update conflicts
        conflicts = self._state.readOverlayConflicts(o.name)
        for layer, forwardPaths in layers.items():
            for forwardPath in forwardPaths:
                vPath = layer.versionedPathFromForward(forwardPath)
                livePath = o.livePathFromVersioned(vPath)
                if util.pathEquals(livePath, forwardPath):
                    self.debug('livePath %s is resolved, unconflicting' %
                        livePath)
                    if livePath in conflicts:
                        conflicts.remove(livePath)
        if conflicts:
            self._state.writeOverlayConflicts(o.name, conflicts)
        else:
            try:
                self._state.deleteOverlayConflicts(o.name)
            except OSError, e:
                if e.errno != errno.ENOENT:
                    raise

        self.debug("%d conflicts left for %r" % (len(conflicts), o))

    def _updateOverlay(self, overlay):
        """
        Update the overlay from the version control system.

        @param overlay: L{savon.overlay.Overlay}
        """
        self._notifyData = []
        overlayPath = self._state.getOverlayDir(overlay.name)

        layersBefore = overlay.getLayerNames()

        self.debug('updating overlay path %r' % overlayPath)
        self._client.update(overlayPath)
        
        for kwargs in self._notifyData:
            path = kwargs['path']
            if path.endswith(
                os.path.join('overlays', overlay.name, 'layers')) \
                and kwargs['action'] == pysvn.wc_notify_action.update_update:
                # layers file got updated; see what happened and act
                layersAfter = overlay.getLayerNames()
                for layerName in layersAfter:
                    if not layerName in layersBefore:
                        self.debug('Updating overlay adds layer %s' %
                            layerName)
                        self._checkoutLayer(overlay.getURL(), layerName)
                        overlay.updateLayers()
                # FIXME: also do the reverse, to find layers that have
                # been removed

    def update(self, livePaths):
        """
        Update everything in the given paths from VCS.
        Will also adjust ownership/permissions from properties.

        @raises savon.util.OverlayConflict: if an overlay is conflicted.

        @rtype:   list of str
        @returns: lines of output
        """
        assert not isinstance(livePaths, str), \
            "update() takes a list, not str"
        self.debug('update(livePaths=%r)' % livePaths)
        res = []

        # find the overlays that manage the given livePaths
        overlaysDict = self.getOverlaysDictForLivePaths(livePaths)

        # do not continue if any overlay is conflicted
        self.debug('update: checking for conflicts')
        for o in overlaysDict.keys():
            conflicts = self._state.readOverlayConflicts(o.name)
            if conflicts:
                raise util.OverlayConflict(o.name, conflicts)

        # update the overlays themselves first
        for o in overlaysDict.keys():
            self._updateOverlay(o)

        # forward all subtrees of the given livePaths before updating from
        # repository
        self.debug('update: forward subtrees of %r' % livePaths)
        self.forwardLayersForOverlaysDict(overlaysDict)
        # we do it again, with recurse is False, which is a cheat to get
        # only the layer -> livePath mapping for exactly the livePaths
        # we specified
        # FIXME: should this be removed ?
        # layersDict = self.forwardLayersForOverlaysDict(overlaysDict,
        #    recurse=False)

        # we should update the vPaths in every layer, but only sync from update
        # to livePath for the top layer

        # make a dict of forwardPath -> overlay, layer, reverse?
        # for the self.svnCallbackNotify to use
        self._forwardPathDict = {}
        for o in overlaysDict.keys():
            self.debug('constructing forwardPath mapping for overlay %r' % o)
            livePaths = overlaysDict[o]
            for livePath in livePaths:
                topLayerToLivePaths = o.findVersionedLivePaths(
                    livePath, onlyTop=True)
                layerToLivePaths = o.findVersionedLivePaths(
                    livePath, onlyTop=False)
                for layer, vLivePaths in layerToLivePaths.items():
                    for vLivePath in vLivePaths:
                        vPath = o.versionedPathFromLive(vLivePath)
                        forwardPath = layer.forwardPathFromVersioned(vPath)
                        # if this forwardPath is not in the top-level layer
                        # for this vPath, then we do not want it forwarded
                        # to the layer
                        reverse = False
                        if vLivePath in topLayerToLivePaths[layer]:
                            reverse = True
                        self._forwardPathDict[forwardPath] = (o, layer, reverse)
                
        # reset the callback's output
        self._notifyOutput = {} # livePath -> list of characters

        # now update all the layers that contain the given livePaths
        for o in overlaysDict.keys():
            livePaths = overlaysDict[o]
            for livePath in livePaths:
                vPath = o.versionedPathFromLive(livePath)
                self.debug('Updating livePath %r, vPath %r' % (livePath, vPath))
                found = False
                for layer in o.layers:
                    # it is possible we're asked to update a livePath that
                    # is not yet in the layer's forward directory
                    # so we need to update everywhere
                    # FIXME: this probably slows everything down a lot;
                    # we can speed it up by checking revision numbers for
                    # each layer that is part of the same repository
                    #if layer.exists(vPath):
                    #    self.debug('vPath %r found in layer %r, updating' % (
                    #        vPath, layer))
                    self._notifyData = []
                    forwardPath = layer.forwardPathFromVersioned(vPath)
                    self.debug('updating forwardPath %r' % forwardPath)
                    # entries = self._client.update(forwardPath)
                    self._client.update(forwardPath)
                    for data in self._notifyData:
                        self.handleNotify(data, o, layer)
                    if layer.exists(vPath):
                        found = True
                if not found:
                    res.append("Path %s not found in any layer." % livePath)
                        
        # collate the output
        livePaths = self._notifyOutput.keys()
        livePaths.sort()
        for livePath in livePaths:
            letters = self._notifyOutput[livePath]
            res.append("%s %s" % ("".join(letters), livePath))

        return res

    def wipe(self):
        self._state.wipe()

    def _updateOverlays(self):
        """
        Update internal cache of overlays based on disk state.
        """
        self._pathToOverlay = {}
        self.overlays = {}
        dirs = glob.glob(os.path.join(self.statedir, 'overlays', '*'))
        if not dirs:
            self.warning('No overlays checked out')
            return {}

        # FIXME: later on we will actually find the correct overlay
        # for now, we only allow one overlay anyway
        if len(dirs) > 1:
            self.error(
                'Savon currently does not support more than one overlay.')

        name = os.path.basename(dirs[0])
        o = overlay.Overlay(name, self.statedir)
        self.debug('_updateOverlays: overlay %r' % o)
        self._pathToOverlay[o.path] = o
        self.overlays[o.name] = o
        
    def getOverlayByPath(self, path):
        """
        Get the overlay that the given path belongs or should belong to.
        Every overlay can be managing a subpart of the system tree.

        @rtype: L{savon.overlay.Overlay}
        """
        self.debug('getOverlayByPath(%s)' % path)
        if not self._pathToOverlay:
            self.error(
                "No overlays found.  Create one with 'create'.")
        self.debug('getOverlayByPath: self._pathToOverlay: %r' %
            self._pathToOverlay)
        for livePath, o in self._pathToOverlay.items():
            # verify that the given path is actually under the overlay's path
            if path.startswith(o.path[:-len(os.path.sep)]):
                return o

        self.error(
            "Path '%s' is not managed by overlay '%s'" % (path, o.name))

    def getOverlayByName(self, name):
        """
        @rtype: L{savon.overlay.Overlay}
        """
        return self.overlays[name]
