# -*- Mode: Python; test-case-name: testsuite.test_overlay -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Savon - sysadmin versioning
# Copyright (C) 2006 Thomas Vander Stichele <thomas at apestaart dot org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.

"""
Wrapper around the svn client to add functionality savon needs.
"""

import sys
import os
import pwd
import grp
import errno

import pysvn
import svn.core

from savon.configure import configure
from savon.util import util, log

properties = ['user', 'group', 'permissions', 'secontext']

class Client(log.Loggable):
    logCategory = 'client'

    # FIXME: add a quiet parameter, default to False
    def __init__(self):
        """
        """
        self.pysvn = pysvn.Client()
        # this gives us a tuple instead of a string, which is more useful
        self.pysvn.exception_style = 1

        self._debug = False

    ### public helper methods
    def mkdir(self, URL, msg):
        self.pysvn.mkdir(URL, msg)

    ### proxy through to pysvn
    def status(self, *args, **kwargs):
        return self.pysvn.status(*args, **kwargs)

    def add(self, *args, **kwargs):
        return self.pysvn.add(*args, **kwargs)

    def revert(self, *args, **kwargs):
        return self.pysvn.revert(*args, **kwargs)

    def resolved(self, *args, **kwargs):
        return self.pysvn.resolved(*args, **kwargs)

    def update(self, *args, **kwargs):
        return self.pysvn.update(*args, **kwargs)

    def isVersioned(self, path):
        """
        Returns whether or not the file is under the VCS's control.
        """
        try:
            entry = self.pysvn.info(path)
            if entry:
                return True
        except pysvn.ClientError:
            pass

        return False

    def getRevision(self, path):
        """
        Returns the current revision of the given path.

        @rtype: int
        """
        entry = self.pysvn.info(path)
        return entry.revision.number
        

    def findRevisionedFiles(self, path):
        """
        Find all the files under version control under the given path.
        This includes files added but not yet commited.
        """
        # we don't use info2, only available with svn >= 1.2.3
        def walker(result, dirname, fnames):
            # check our dirname first; don't descend if it's not versioned
            if not self.isVersioned(dirname):
                # just assigning [] creates a new var, so doesn't work
                # fnames = []

                # there is probably a nicer way, but hey ...
                while fnames:
                    del fnames[0]
                return
            
            for fname in fnames:
                fullpath = os.path.join(dirname, fname)
                if self.isVersioned(fullpath):
                    result.append(fullpath)

        result = []
        if self.isVersioned(path):
            result.append(path)

        os.path.walk(path, walker, result)

        return result

    def getSavonProperties(self, path):
        """
        Get the local savon properties on the given .savon path as a dictionary.

        @rtype: dict of str -> str
        """
        d = {}
        for key in properties:
            d[key] = None
            
        props = self.pysvn.proplist(path)
        if not props: return d
            
        for k, v in props[0][1].items():
            if k.startswith('savon:'):
                # svn pe adds a \n even when we don't want to, so handle
                # that case gracefully with rstrip
                d[k[6:]] = v.rstrip()

        return d

    def setSavonProperty(self, path, property, value):
        """
        Set the savon property to the given value for the given file.
        """
        assert property in properties
        if not value:
            self.warning("No value to set on property %s on %s" % (
                property, path))
            return
        key = 'savon:%s' % property
        self.debug('Setting %s to %s on %s' % (key, value, path))
        self.pysvn.propset(key, value + "\n", path)

    def diffSavonProperties(self, path):
        """
        Return a dict of savon properties that are different between the
        current file status and the local svn properties.
        The current file value is returned as the value.
        """
        # symlinks have their own properties
        s = os.lstat(path)
        value = util.getPermissions(s)

        user = None
        group = None

        try:
            user = pwd.getpwuid(s.st_uid)[0]
        except KeyError:
            self.warning('No uid %d found in passwd' % s.st_uid)

        try:
            group = grp.getgrgid(s.st_gid)[0]
        except KeyError:
            self.warning('No gid %d found in groups' % s.st_gid)

        current = { 
          'permissions': oct(value),
          'user':        user,
          'group':       group,
          'secontext':   None,
        }
        # a symlink does not have an secontext of its own
        if configure.selinux and not os.path.islink(path):
            context = configure.selinux.getfilecon(path)
            current['secontext'] = context[1]
            self.log('diffSavonProperties: path %s: current context %r' % (
                path, context))

        d = self.getSavonProperties(path)
        self.log('diffSavonProperties: path %s: savon props %r' % (path, d))

        ret = {}

        for k in properties:
            if d[k] != current[k]:
                ret[k] = current[k]
                
        return ret

    def updateSavonProperties(self, path):
        """
        Update savon properties on the given path if needed; setting them
        to the values the path currently has.

        Call this before comparing to the VCS version.
        """
        d = self.diffSavonProperties(path)

        for k in properties:
            if d.has_key(k):
                self.setSavonProperty(path, k, d[k])

    def updateFromSavonProperties(self, path):
        """
        Update from savon properties on the given .savon path if needed;
        setting the path to the values from VCS.
        """
        vcs = self.getSavonProperties(path)
        diffs = self.diffSavonProperties(path)
        self.log('path has diffs in %r' % diffs.keys())

        if diffs.has_key('user'):
            user = vcs['user']
            if user is None:
                self.error('Path %s does not have savon:user set' % path)
            try:
                os.lchown(path, pwd.getpwnam(user)[2], -1)
                self.debug("user of %s changed to %s" % (path, user))
            except KeyError:
                self.warning('Did not find user %s in passwd' % user)
            except OSError:
                self.warning('Could not change user of %s from %s to %s' % (
                    path, diffs['user'], user))
        if diffs.has_key('group'):
            group = vcs['group']
            if group is None:
                self.error("No group information on %s" % path)
            try:
                os.lchown(path, -1, grp.getgrnam(group)[2])
                self.debug("group of %s changed from %s to %s" % (
                    path, diffs['group'], group))
            except OSError:
                self.warning('Could not change group of %s to %s' % (
                    path, group))
        if diffs.has_key('permissions'):
            permissions = vcs['permissions']
            os.chmod(path, int(permissions, 8))
            self.debug("permissions of %s changed from %s to %s" % (
                path, diffs['permissions'], permissions))

        # handle secontext changes
        if diffs.has_key('secontext'):
            context = vcs['secontext']
            try:
                self._setSEContext(path, diffs['secontext'], context)
            except OSError, e:
                if e.errno == errno.EINVAL:
                    # try again with :s0 if applicable
                    if not context.endswith(':s0'):
                        self._setSEContext(path, diffs['secontext'],
                            context + ':s0')
                    else:
                        sys.stderr.write(
                            'WARNING: Could not set secontext %s on %s\n' % (
                                context, path))

    def _setSEContext(self, path, fromContext, context):
        # set SEContext if we can

        self.debug("changing security context of %s from %s to %s" % (
            path, fromContext, context))

        if not configure.selinux:
            self.debug('cannot set selinux context')
            return

        if not context:
            self.warning('No secontext property in repository for %s' %
                path)
            return

        configure.selinux.setfilecon(path, context)

    def existsURL(self, URL):
        """
        Check if the given URL exists.

        @rtype: bool
        """
        head = pysvn.Revision(pysvn.opt_revision_kind.head)
        try:
            # entry = self.pysvn.info2(URL, revision=head)
            self.pysvn.info2(URL, revision=head)
            return True
        except pysvn._pysvn.ClientError, e:
            message, code = e.args[1][0]
            if code == svn.core.SVN_ERR_RA_ILLEGAL_URL:
                return False
            elif code == svn.core.SVN_ERR_ENTRY_MISSING_URL:
                self.error("'%s' is not a URL." % URL)
            else:
                self.error("Subversion error: %s" % message)
