# -*- Mode: Python -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Savon - sysadmin versioning
# Copyright (C) 2006 Thomas Vander Stichele <thomas at apestaart dot org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.

"""
'overlay' command.
"""

import sys

from savon import actor
from savon.util import log, util


class Add(util.LogCommand):
    usage = "add [URL] [layer name]"
    description = "add a layer to the given overlay, below the current one"

    def do(self, args):
        self.debug('do(%s)', args)
        try:
            URL = args[0]
            name = args[1]
        except IndexError:
            self.outputUsage()
            return 1

        # overlay name
        overlayName = self.parentCommand.getOverlayName()
        actor = self.getRootCommand().actor
        return actor.addLayer(URL, overlayName, name)

class Down(util.LogCommand):
    usage = "down"
    description = "move one layer down in the stack of layers"

    def do(self, args):
        self.debug('do(%s)', args)
        overlay = self.parentCommand.getOverlay()
        try:
            layer = overlay.down()
        except IndexError:
            self.stderr.write('Already at bottom layer.\n')
            return 1
        self.stdout.write("Now at layer '%s'.\n" % layer)

class Up(util.LogCommand):
    usage = "up"
    description = "move one layer up in the stack of layers"

    def do(self, args):
        self.debug('do(%s)', args)
        overlay = self.parentCommand.getOverlay()
        try:
            layer = overlay.up()
        except ValueError:
            self.stderr.write('Already at top layer.\n')
            return 1

        if not layer:
            self.stdout.write('On top of the stack, no layer.\n')
        else:
            self.stdout.write("Now at layer '%s'.\n" % layer)

class Go(util.LogCommand):
    summary = "go directly to the given layer"

    def do(self, args):
        self.debug('do(%s)', args)
        if not args:
            self.stderr.write('Specify a layer to go to.\n')
            return 1

        overlay = self.parentCommand.getOverlay()
        layerName = args[0]
        try:
            overlay.go(layerName)
        except KeyError:
            self.stderr.write("No layer '%s' in overlay.\n")
            return 1
        self.stdout.write("Now at layer '%s'.\n" % layerName)


class Status(util.LogCommand):
    usage = "status"
    description = "show status of overlay"

    def do(self, args):
        self.debug('do(%s)', args)
        # overlay name
        name = self.parentCommand.overlayName
        actor = self.getRootCommand().actor
        names = actor.overlays.keys()
        self.debug('overlay names: %r' % names)
        names.sort()

        found = False
        for n in names:
            o = actor.getOverlayByName(n)
            if name and o.name != name:
                continue
            found = True
            self.stdout.write('Overlay %s:\n' % o.name)
            self.stdout.write('- path %s\n' % o.path)
            self.stdout.write('- layers:\n')
            for l in o.layers:
                name = l.name
                if l == o.current:
                    name += ' (current)'
                self.stdout.write('  - %s\n' % name)

        if not name:
            self.stderr.write("No overlays found.\n"
                "Create one with 'savon create'.\n")
            return 1

        if not found:
            self.stderr.write("No overlay '%s' found.\n" % name)
            return 1

class Wipe(util.LogCommand):
    summary = "wipe out savon state"
    description = """Remove savon's state for the overlay.
Used when you want to stop managing a machine using savon, or want to
do a fresh checkout of an overlay."""

    def do(self, args):
        self.debug('do(%s)', args)
        actor = self.getRootCommand().actor
        actor.wipe()

class Overlay(util.LogCommand):
    usage = "overlay [overlay-options] %command"
    description = "manage overlays"
    subCommandClasses = (Add, Down, Up, Go, Status, Wipe)

    def addOptions(self):
        self.debug('adding option')
        self.parser.add_option('-n', '--name',
            action="store", dest="name",
            help="name of overlay to operate on")

    def handleOptions(self, options):
        self.debug('handling options %r' % options)
        self.overlayName = options.name

    def getOverlayName(self):
        """
        Return the chosen overlay name, or the default one if not chosen.

        @rtype: str or None
        """
        if self.overlayName:
            return self.overlayName

        actor = self.getRootCommand().actor
        names = actor.overlays.keys()
        if len(names) > 1:
            sys.stderr.write('Please specify an overlay with -n [name].\n')
            return None

        return names[0]

    def getOverlay(self):
        """
        Return the chosen overlay, or the default one if not chosen.

        @rtype: L{savon.overlay.Overlay} or None
        """
        name = self.getOverlayName()
        if not name:
            return None

        actor = self.getRootCommand().actor
        return actor.overlays[name]
