# -*- Mode: Python; test-case-name: savon.test.test_layer -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Savon - sysadmin versioning
# Copyright (C) 2006 Thomas Vander Stichele <thomas at apestaart dot org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.

"""
Layer
"""

import os

from savon import client
from savon.util import util, log

class Layer(log.Loggable):
    """
    @ivar name:    the name of the layer
    @type name:    str
    @ivar dir:     absolute path to this layer's state directory;
                   inherited from the overlay this layer object is created for.
    """
    logCategory = 'layer'

    def __init__(self, name, statedir=None):
        """
        @param name:     name of the layer
        @param statedir: savon state directory
                         defaults to $HOME/.savon
        """
        self.debug("Creating Layer with name '%s'" % name)
        self._client = client.Client()
        self.name = name
        self._statedir = statedir
        if not self._statedir:
            home = os.path.expanduser('~')
            self._statedir = os.path.join(home, '.savon')

        self.dir = os.path.join(self._statedir, 'layers', name)
        if not os.path.exists(self.dir):
            self.error("Could not find layer '%s' in %s" % (name, self.dir))

    def __repr__(self):
        return "<Layer %s>" % self.name

    def exists(self, vPath):
        """
        Check if the given versioned path is present in the layer's forward
        checkout.
        
        @param vPath: a versioned path
        @type  vPath: str

        @rtype: bool
        """
        assert os.path.isabs(vPath), \
            "Internal error: versioned path %s is not absolute" % vPath
        suffix = vPath
        if vPath is os.path.sep:
            suffix = ''
        # broken symlinks are still considered existing, since they can be
        # pointing to a file in another layer
        if os.path.lexists(
            os.path.join(self.dir, 'forward') + suffix):
            self.debug('versioned path %r exists in layer %r' % (vPath,
                self.name))
            return True
        return False

    def forward(self, livePath, vPath):
        """
        'Forward' the given livePath to the given versioned path.

        This copies the live path, preserving attributes.

        @returns: the full path this livePath was forwarded to, so it can
                  be commited
        """
        self.debug('forward(%s, %s)' % (livePath, vPath))
        # broken symlinks are still considered versionable
        if not os.path.lexists(livePath):
            self.error("Path '%s' does not exist." % (livePath))
        targetPath = self.forwardPathFromVersioned(vPath)

        self.debug('mirroring %s to %s' % (livePath, targetPath))
        try:
            util.mirror(livePath, targetPath)
        except util.ParentDoesNotExist:
            raise util.ParentNotVersioned(os.path.dirname(livePath))
        return targetPath

    def reverse(self, livePath, vPath):
        """
        'Reverse' the given livePath from the given versioned path.

        This copies from the repository path, preserving attributes.

        @returns: the full path this livePath was reversed from
        """
        self.debug('reverse(%s, %s)' % (livePath, vPath))
        forwardPath = self.forwardPathFromVersioned(vPath)

        self.debug('mirroring %s to %s' % (forwardPath, livePath))
        util.mirror(forwardPath, livePath)
        return forwardPath

    def forwardPathFromVersioned(self, vPath):
        """
        Given the versioned path, return the full path to this path in
        the forward staging area.

        For overlay of /:
        /     : /    -> $HOME/.savon/layers/[layer name]/forward
        /etc  : /etc -> $HOME/.savon/layers/[layer name]/forward/etc

        For overlay of /home:
        /home : /    -> $HOME/.savon/layers/[layer name]/forward

        @type vPath: str

        @rtype: str
        """
        assert os.path.isabs(vPath), \
            "Internal error: vPath %s is not absolute" % vPath
        if vPath == os.path.sep:
            return os.path.join(self.dir, 'forward')
        return os.path.join(self.dir, 'forward') + vPath

    def findVersionedPaths(self, vPath):
        """
        Find all the paths underneath vPath that are versioned in this
        layer.

        @return: a list of versioned paths
        """
        self.debug('findVersionedPaths(%s)' % vPath)
        def walker(result, dirname, fnames):
            # check our dirname first; don't descend if it's not versioned
            if not self._client.isVersioned(dirname):
                # just assigning [] creates a new var, so doesn't work
                # fnames = []

                # there is probably a nicer way, but hey ...
                while fnames:
                    del fnames[0]
                return
            
            result.append(self.versionedPathFromForward(dirname))

            for fname in fnames:
                forwardPath = os.path.join(dirname, fname)
                if self._client.isVersioned(forwardPath):
                    result.append(self.versionedPathFromForward(forwardPath))

        result = []
        forwardPath = self.forwardPathFromVersioned(vPath)
        self.debug("walking versioned path '%s' which is forward path '%s'" % 
            (vPath, forwardPath))
        os.path.walk(forwardPath, walker, result)

        # after walking, check this path ourselves
        if self._client.isVersioned(forwardPath):
            vPath = self.versionedPathFromForward(forwardPath)
            if vPath not in result:
                result.append(vPath)

        result = util.unique(result)
        self.debug('findVersionedPaths(%s): %r' % (vPath, result))
        return result

    def versionedPathFromForward(self, forwardPath):
        """
        Given the forward staging area path, return the versioned path.
        .../forward     -> /
        .../forward/etc -> /etc
        """
        f = os.path.join(self.dir, 'forward')
        assert forwardPath.startswith(f), \
            "Internal error: path %s is not a forwardPath" % forwardPath
        ret = forwardPath[len(f):]
        if ret == '':
            ret = os.path.sep
        return ret
