# -*- Mode: Python; test-case-name: testsuite.test_savon -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Savon - sysadmin versioning
# Copyright (C) 2006 Thomas Vander Stichele <thomas at apestaart dot org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.

"""
Implementation of the savon program.
"""

import sys

import svn
import pysvn

from savon import actor
from savon.util import log, util

from savon.command import overlay, resolve

def main(argv,stdout=sys.stdout, stderr=sys.stderr, actor=None):
    """
    @param stdout: the standard output file object to use
    @param stderr: the standard error file object to use
    @param actor:  an actor to use, which is mostly useful for the testsuite
    """
    # if an actor is specified, we use its stdout and stderr
    if actor:
        Savon.actor = actor
        stdout = actor.stdout
        stderr = actor.stderr

    c = Savon(stdout=stdout, stderr=stderr)
    try:
        ret = c.parse(argv)
    except SystemError, e:
        stderr.write('savon: error: %s\n' % e.args)
        return 255

    if ret is None:
        return 0
    return ret

# these are all the commands that are too small to go in their own file
class Add(util.LogCommand):
    usage = "add PATH..."
    description = "add the given paths to savon for managing"

    def do(self, paths):
        paths = util.getAbsolutePaths(paths)
        if not paths:
            self.stderr.write("Please specify paths to add.\n")
            return 1
        actor = self.getRootCommand().actor
        self.info('Adding paths %r' % paths)
        try:
            output = actor.add(paths)
        except util.ParentNotVersioned, e:
            self.warning('parent path %s not under version control' % e.args)
            self.stderr.write("savon: '%s' is not a working copy\n" % e.args)
            return 1
        except pysvn._pysvn.ClientError, e:
            message, code = e.args[1][0]
            if code == svn.core.SVN_ERR_ENTRY_EXISTS:
                self.stderr.write("savon: %s\n" % message)
                return 1
            else:
                self.error("Subversion error: %s\n" % message)

        if not output:
            return
        for line in output:
            self.stdout.write('%s\n' % line)


class Checkout(util.LogCommand):
    usage = "checkout [URL] [overlay name]"
    description = "check out the given overlay"
    aliases = ("co", )

    def do(self, args):
        self.debug('do(%r)', args)
        try:
            URL = args[0]
            name = args[1]
        except IndexError:
            self.outputUsage()
            return 1

        actor = self.getRootCommand().actor
        conflicts = actor.checkout(URL, name, force=self.force,
            livePath=self.livePath)
        if conflicts:
            self.stdout.write('The following paths are in conflict:\n')
            for livePath in conflicts:
                self.stdout.write('  %s\n' % livePath)

    def addOptions(self):
        self.parser.add_option('-f', '--force',
            action="store_true", dest="force", default=False,
            help="force checkout by preferring the repository version")
        self.parser.add_option('-l', '--live-path',
            action="store_true", dest="live_path", default=False,
            help="manually override the live path managed by the overlay")

    def handleOptions(self, options):
        self.force = options.force
        self.livePath = options.live_path

class Commit(util.LogCommand):
    usage = "commit [list of live paths]"
    description = "commit the trees under the given live paths"

    def do(self, paths):
        self.debug('do(%r)', paths)
        paths = util.getAbsolutePaths(paths)
        actor = self.getRootCommand().actor
        actor.commit(paths, message=self._message, layer=self._layer)

    def handleOptions(self, options):
        self._message = options.message
        self._layer = options.layer

    def addOptions(self):
        self.parser.add_option('-m', '--message',
            action="store", dest="message",
            help="commit message")
        self.parser.add_option('-l', '--layer',
            action="store", dest="layer",
            help="commit only changes to the given layer")

class Create(util.LogCommand):
    usage = "create [URL] [overlay name] [path]"
    description = "create a savon overlay at the given URL for the given path"

    def do(self, args):
        self.debug('do(%r)', args)
        try:
            URL = args[0]
            name = args[1]
        except IndexError:
            self.outputUsage()
            return 1
        rest = ()
        if len(args) > 2:
            rest = (args[2], )
        actor = self.getRootCommand().actor
        self.debug('actor.create()')
        actor.create(URL, name, *rest)
 
class Diff(util.LogCommand):
    usage = "diff [live path]"
    description = "show the diff for the given live path"

    def do(self, args):
        self.debug('do(%r)', args)
        actor = self.getRootCommand().actor
        paths = util.getAbsolutePaths(args)
        output = actor.diff(paths)
        for line in output:
            self.stdout.write("%s\n" % line)

class Info(util.LogCommand):
    usage = "info [live path]"
    description = "show repository information about the overlay"

    def do(self, args):
        self.debug('do(%r)', args)
        actor = self.getRootCommand().actor
        paths = util.getAbsolutePaths(args)
        output = actor.information(paths[0])
        for line in output:
            self.stdout.write("%s\n" % line)

class Revert(util.LogCommand):
    usage = "revert [live path]"
    description = "revert the live path back to the repository version"

    def do(self, args):
        self.debug('do(%r)', args)
        actor = self.getRootCommand().actor
        paths = util.getAbsolutePaths(args)
        output = actor.revert(paths)

        for line in output:
            self.stdout.write("%s\n" % line)

class Status(util.LogCommand):
    usage = "status [list of live paths]"
    description = "show the status of the trees under the given live paths"
    aliases = ('st', 'stat')

    def addOptions(self):
        self.parser.add_option('-v', '--verbose',
                          action="store_true", dest="verbose",
                          default=False,
                          help="print extra information")
        self.parser.add_option('-N', '--non-recursive',
                          action="store_false", dest="recursive",
                          default=True,
                          help="operate on single directory only")
        self.parser.add_option('-q', '--quiet',
                          action="store_true", dest="quiet",
                          default=False,
                          help="print as little as possible")
 
    def handleOptions(self, options):
        self.options = options

    def do(self, paths):
        self.debug('do(%r)', paths)
        paths = util.getAbsolutePaths(paths)
        actor = self.getRootCommand().actor
        output = actor.status(paths, recurse=self.options.recursive,
            quiet=self.options.quiet)
        layerNameLen = 0
        for livePath, layerName, versioned, status, revision, \
            commit_author, commit_revision in output:
            # first find longest layerName
            if len(layerName) > layerNameLen:
                layerNameLen = len(layerName)
            
        for livePath, layerName, versioned, status, revision, \
            commit_author, commit_revision in output:
            if not self.options.verbose:
                # do not show lines that do not have status, ie are uptodate
                if status.rstrip():
                    formatString = "%s %" + str(layerNameLen) + "s %s\n"
                    self.stdout.write(formatString % (
                        status, layerName, livePath))
            else:
                if versioned:
                    formatString = "%s  %8s %8s %-12s %" + str(layerNameLen) \
                        + "s %s\n"
                    self.stdout.write(formatString % (
                        status, revision, commit_revision, commit_author,
                        layerName, livePath))
                else:
                    formatString = "%s  %8s %8s %-12s %" + str(layerNameLen) \
                        + "s %s\n"
                    self.stdout.write(formatString % (
                        status, '', '', '', layerName, livePath))


class Update(util.LogCommand):
    usage = "update [list of live paths]"
    description = "update the list of live paths from the repository"
    aliases = ('up', )

    def do(self, args):
        self.debug('do(%r)', args)
        actor = self.getRootCommand().actor
        paths = util.getAbsolutePaths(args)
        try:
            output = actor.update(paths)
        except util.OverlayConflict, e:
            overlayName, conflicts = e.args
            self.stderr.write("Overlay '%s' has the following conflicts:\n" %
                overlayName)
            for c in conflicts:
              self.stderr.write("  %s\n" % c)
            self.stderr.write("\nPlease use 'savon resolve' "
                "to resolve conflicts before updating.\n")
            return 1
        for line in output:
            self.stdout.write("%s\n" % line)

class Savon(util.LogCommand):
    """
    @cvar actor: the actor that will execute for us
    @type actor: L{actor.Actor}
    """
    usage = "savon %command"
    description = """Savon stores part of your file tree in subversion,
allowing you to save history of your configuration data.  It keeps track of
ownership and permissions."""
    subCommandClasses = [
        Add, Checkout, Commit, Create, Diff, Info,
        overlay.Overlay, resolve.Resolve, Revert, Status, Update
    ]
    actor = None

    def addOptions(self):
        self.parser.add_option('-d', '--debug',
                          action="store_true", dest="debug",
                          default=False,
                          help="show debug information")
        self.parser.add_option('', '--version',
                               action="store_true", dest="version",
                               default=False,
                               help="show version information")

        # FIXME: is this the right place ?
        log.init()

        # We allow setting it as a class var to make testing easier,
        # but in the normal case we will have to create our own actor 
        if not self.actor:
            self.actor = actor.Actor(stdout=self.stdout, stderr=self.stderr)
        return

    def handleOptions(self, options):
        if options.version:
            from savon.configure import configure
            print "savon %s" % configure.version
            sys.exit(0)
