# -*- Mode: Python; test-case-name: testsuite.test_overlay -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Savon - sysadmin versioning
# Copyright (C) 2006 Thomas Vander Stichele <thomas at apestaart dot org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.

"""
Overlay
"""

import os
import commands

from savon import layer, client, state
from savon.util import log, util

class Overlay(log.Loggable):
    """
    I manage the overlay.
    An overlay consists of the path that the overlay manages,
    and a list of layers that the overlay consists of.

    The layers are stacked bottom to top; a file in a higher layer 'wins'
    from a file in a lower layer.

    I do not interact directly with the version control backend.
    I only wrap the state of the overlay as it exists on-disk in the state
    directory.

    @ivar name:    the name of the overlay
    @type name:    str
    @ivar path:    path this is an overlay for
                   (does not end with a separator, except for /)
                   If the path is overridden, this will be the overridden path
    @ivar layers:  list of layers defined for this overlay, from top to bottom
    @type layers:  list of L{savon.layer.Layer}
    @ivar current: current layer
    @type current: L{savon.layer.Layer}
    """

    logCategory = 'overlay'
    current = None
    path = None

    def __init__(self, name, statedir=None):
        """
        @param name:     name of the overlay
        @param statedir: savon state directory
                         defaults to $HOME/.savon
        """
        self.name = name
        self._state = state.State(statedir)
        self.debug('Creating Overlay %s from state dir %s' % (
            name, self._state.dir))

        if not self._state.existsOverlay(name):
            self.error('Could not find overlay')

        self.path = self._state.readOverlayPath(name)
        if self.path is not os.path.sep and self.path.endswith(os.path.sep):
            self.error("Path '%s' should not end in %s'" % (
                self.path, os.path.sep))
        self.debug('I am an overlay for path %s' % self.path)

        self.updateLayers()
        self._client = client.Client()

    def __repr__(self):
        return "<Overlay %s for path %s>" % (self.name, self.path)

    def updateLayers(self):
        """
        Update my idea of which layers are in the overlay.
        Layers can be added to an overlay due to an update.
        """
        self.layers = []

        currentName = self._state.readOverlayCurrent(self.name)
        layerNames = self.getLayerNames()
        self.debug('current layer %r, all layers %r' % (
            currentName, layerNames))

        for name in layerNames:
            self.debug("appending layer for name '%s'" % name)
            l = layer.Layer(name, statedir=self._state.dir)
            self.layers.append(l)
            if name == currentName:
                self.current = l

    def getLayerNames(self):
        """
        Return a list of layer names in this overlay.

        @rtype: list of str
        """
        return self._state.readOverlayLayers(self.name)

    def down(self):
        """
        Move one layer down in the layer stack.

        @raises IndexError: if we are already at the bottom of the stack

        @rtype:   str
        @returns: name of new layer
        """
        if not self.current:
            # we are above the first layer, so move to first layer
            self.current = self.layers[0]
        else:
            i = self.layers.index(self.current)
            if i + 1 == len(self.layers):
                raise IndexError

            self.current = self.layers[i + 1]

        self._state.writeOverlayCurrent(self.name, self.current.name)

        return self.current.name

    def up(self):
        """
        Move one layer up in the layer stack.

        @raises ValueError: if we are already at the top of the stack

        @rtype:   str or None
        @returns: name of new layer if we moved or None if we're at the top
                  of the layer stack.
        """
        i = self.layers.index(self.current)

        if i == 0:
            self.current = None
        else:
            self.current = self.layers[i - 1]

        name = None
        if self.current:
            name = self.current.name
        self._state.writeOverlayCurrent(self.name, name)
        return name

    def go(self, layerName):
        """
        Go to the given layer in the layer stack.

        @raises KeyError: if the given layer name is not in the overlay.
        """
        for layer in self.layers:
            if layer.name == layerName:
                self._state.writeOverlayCurrent(self.name, layerName)
                self.debug('Moved to layer %s' % layerName)
                return

        raise KeyError, layerName

    def exists(self, livePath):
        """
        Check if the given livePath is present (ie. versioned) in the
        overlay's forward checkouts.
        
        @param livePath: a path on the live system
        @type  livePath: str

        @return: the top-most layer this file exists in
        @rtype:  L{savon.layer.Layer}
        """
        self.log('exists(%s)' % livePath)
        assert os.path.isabs(livePath)
        # first strip off the overlay's path
        assert livePath.startswith(self.path), \
            "livePath %s should start with overlay path %s" % (
                livePath, self.path)
        vPath = self.versionedPathFromLive(livePath)

        self.log('Looking for versioned path %s' % vPath)
        for layer in self.layers:
            if layer.exists(vPath):
                self.debug('Found versioned path %r in layer %r' % (
                    vPath, layer))
                return layer
        self.debug('vPath %s not in any layer' % vPath)
        return None

    def add(self, livePath):
        """
        Add the given path to the current layer in the overlay.
        """
        self.debug('Adding live path %s' % livePath)
        fullPath = self.forward(livePath, layer=self.current)
        self._client.add([fullPath, ], recurse=False)
        self.updateSavonProperties(livePath, layer=self.current)

    def getURL(self):
        """
        Return the URL of the savon repository this overlay lives in.
        """
        overlayDir = self._state.getOverlayDir(self.name)
        (status, output) = commands.getstatusoutput('svn info %s' % overlayDir)
        url = None
        for line in output.split('\n'):
            if line.startswith('URL: '):
                parts = line[len('URL: '):].split('/')
                # the url ends in /overlays/(name), which we need to chop off
                url = ('/').join(parts[:-2])

        return url

    def forward(self, livePath, layer=None):
        """
        'Forward' the given livePath into its layer.

        @type  livePath: str
        @param layer:    the layer to forward this path to

        @return: the full forward path where this livePath was forwarded to.
        """
        assert os.path.isabs(livePath)
        self.debug('forward(%s, layer=%r)' % (livePath, layer))
        if not layer:
            layer = self.exists(livePath)
        if not layer:
            self.error("Path '%s' is not in the overlay's path '%s'." % (
                livePath, self.path))

        return layer.forward(livePath, self.versionedPathFromLive(livePath))

    def reverse(self, livePath, layer=None):
        """
        'Reverse' the given livePath from its layer.

        @type  livePath: str
        @param layer:    the layer to reverse this path from, or None
                         to find the layer automatically.

        @return: the full path where this livePath was reversed from.
        """
        assert os.path.isabs(livePath)
        self.debug('reverse(%s, layer=%r)' % (livePath, layer))
        if not layer:
            layer = self.exists(livePath)
        if not layer:
            self.error("Path '%s' is not in the overlay's path '%s'." % (
                livePath, self.path))

        return layer.reverse(livePath, self.versionedPathFromLive(livePath))

    def updateFromSavonProperties(self, livePath, layer):
        """
        Update from savon properties for the given livePath in the given layer.
        """
        self.debug('updateFromSavonProperties(%s, %r)' % (livePath, layer))
        forwardPath = layer.forwardPathFromVersioned(
            self.versionedPathFromLive(livePath))
        self._client.updateFromSavonProperties(forwardPath)

    def updateSavonProperties(self, livePath, layer):
        """
        Update savon properties for the given livePath to the given layer.
        """
        self.debug('updateSavonProperties(%s, %r)' % (livePath, layer))
        # find the forwardPath in the layer's forward area
        forwardPath = layer.forwardPathFromVersioned(
            self.versionedPathFromLive(livePath))
        self._client.updateSavonProperties(forwardPath)

    def findVersionedLivePaths(self, livePath, onlyTop=True):
        """
        Find all files under the given livePath that are versioned in
        one of our layers.
        If onlyTop, then for every such path, find only the top-most layer.

        @param onlyTop: whether to only return the top layer for each path.
        @rtype:         a dict of layer -> list of livePath
        """
        self.debug('findVersionedLivePaths(%r, onlyTop=%r)' % (
            livePath, onlyTop))
        vPathsFound = [] # so we can track what we already found
        ret = {}
        self.debug('%d layers' % len(self.layers))
        for l in self.layers:
            self.debug('Finding versioned paths for layer %r' % l)
            vPaths = l.findVersionedPaths(self.versionedPathFromLive(livePath))
            if onlyTop:
                vPaths = [p for p in vPaths if p not in vPathsFound]
            vPathsFound.extend(vPaths)
            # make paths live
            ret[l] = util.unique(
                [self.livePathFromVersioned(p) for p in vPaths])
            # sort them, so that parent directories come before their children
            ret[l].sort()
        self.debug('findVersionedLivePaths(%r): %d layers, %d files' % (
            livePath, len(ret), len(vPathsFound)))
        return ret

    def versionedPathFromLive(self, livePath):
        """
        Given a live path, return the versioned path.

        overlay /:     /            -> /
        overlay /:     /etc         -> /etc
        overlay /home: /home/thomas -> /thomas

        @rtype: str
        """
        # what we want to remove from livePath
        assert os.path.isabs(livePath), \
            "Internal error: livePath %s should be absolute" % livePath

        if livePath == self.path:
            return os.path.sep

        suffix = livePath
        if self.path is not os.path.sep:
            assert livePath.startswith(self.path), \
                "Internal error: livePath %s should start with %s" % (
                    livePath, self.path)
            suffix = livePath[len(self.path):]

        return suffix

    def livePathFromVersioned(self, vPath):
        """
        Given a versioned path, return the live path.

        overlay /: /           -> /
        overlay /: /etc        -> /etc
        overlay /home: /       -> /home
        overlay /home: /thomas -> /home/thomas

        @rtype: str
        """
        assert os.path.isabs(vPath), "versioned path %s is not absolute" % vPath
        if vPath == os.path.sep:
            return self.path
        if self.path == os.path.sep:
            return vPath
        return self.path + vPath

    def order(self, layer1, layer2):
        """
        Check whether layer1 is above layer2 or not.

        @raises IndexError: if the overlay does not have layer1 or layer2
        @rtype: int
        @returns: -1, 0 or 1 if layer1 is below, same as, or above layer2
        """
        i1 = self.layers.index(layer1)
        i2 = self.layers.index(layer2)
        if i1 < i2: return 1
        if i1 > i2: return -1
        if i1 == i2: return 0
