# -*- Mode: Python; test-case-name: testsuite.test_actor -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Savon - sysadmin versioning
# Copyright (C) 2006 Thomas Vander Stichele <thomas at apestaart dot org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.

"""
The state object manages the local savon state.
"""

import errno
import os

from savon.util import log

class State(log.Loggable):
    def __init__(self, dir=None):
        """
        @param dir: savon state directory
                    defaults to $HOME/.savon
        """
        self.dir = dir
        if not self.dir:
            home = os.path.expanduser('~')
            self.dir = os.path.join(home, '.savon')

    def getOverlayDir(self, overlayName):
        assert overlayName, "No overlayName given"
        return os.path.join(self.dir, 'overlays', overlayName)

    def getLayerForwardDir(self, layerName):
        return os.path.join(self.dir, 'layers', layerName, 'forward')

    def _deleteOverlayFile(self, overlayName, fileName):
        localOverlayDir = self.getOverlayDir(overlayName)
        os.unlink(os.path.join(localOverlayDir, fileName))

    def _readOverlayFile(self, overlayName, fileName):
        localOverlayDir = self.getOverlayDir(overlayName)
        handle = open(os.path.join(localOverlayDir, fileName), 'r')
        data = handle.read()
        return data

    def _writeOverlayFile(self, overlayName, fileName, data):
        localOverlayDir = self.getOverlayDir(overlayName)
        handle = open(os.path.join(localOverlayDir, fileName), 'w')
        if data:
            handle.write(data)
        handle.close()

    def deleteOverlayConflicts(self, overlayName):
        self._deleteOverlayFile(overlayName, 'conflicts')

    def appendOverlayConflicts(self, overlayName, livePaths):
        data = None
        if livePaths:
            data = "\n".join(livePaths) + "\n"
        localOverlayDir = self.getOverlayDir(overlayName)
        handle = open(os.path.join(localOverlayDir, 'conflicts'), 'a')
        if data:
            handle.write(data)
        handle.close()

    def readOverlayConflicts(self, overlayName):
        """
        Returns an empty list if there are no conflicts.

        @returns: list of paths to nodes that are in conflict
        @rtype:   list of str
        """
        livePaths = []

        try:
            data = self._readOverlayFile(overlayName, 'conflicts')
        except IOError, e:
            if e.errno == errno.ENOENT:
                return livePaths
            raise

        if data:
            livePaths = data.rstrip().split('\n')
        return livePaths

    def writeOverlayConflicts(self, overlayName, livePaths):
        data = None
        if livePaths:
            data = "\n".join(livePaths) + "\n"
        self._writeOverlayFile(overlayName, 'conflicts', data)

    def readOverlayCurrent(self, overlayName):
        """
        @returns: current layer name
        """
        self.debug('read current layer for overlayName %r' % overlayName)
        data = self._readOverlayFile(overlayName, 'current')
        current = data.rstrip()
        if not current:
            self.debug('None current layer')
            return None

        self.debug('Current layer: %r' % current)
        return current

    def writeOverlayCurrent(self, overlayName, layerName):
        self.debug('writing current layer %s for overlay %s' % (
            layerName, overlayName))
        self._writeOverlayFile(overlayName, "current", "%s\n" % layerName)

    def readOverlayLayers(self, overlayName):
        """
        @returns: list of names of layers in this overlay
        @rtype:   list of str
        """
        try:
            data = self._readOverlayFile(overlayName, 'layers')
        except IOError:
            self.error('Repository problem: could not read overlay layers file')
            return None
        layerNames = []
        if data:
            layerNames = data.rstrip().split('\n')
        return layerNames

    def writeOverlayLayers(self, overlayName, layerNames):
        data = None
        if layerNames:
            data = "\n".join(layerNames) + "\n"
        self._writeOverlayFile(overlayName, 'layers', data)

    def readOverlayPath(self, overlayName):
        """
        Get the path managed by the overlay.  If the path is overridden,
        return the override path.

        @returns: path managed by the overlay
        """
        try:
            data = self._readOverlayFile(overlayName, 'path.override')
        except IOError:
            data = self._readOverlayFile(overlayName, 'path')

        return data.rstrip()

    def writeOverlayPath(self, overlayName, path):
        self._writeOverlayFile(overlayName, "path", "%s\n" % path)

    def writeOverrideOverlayPath(self, overlayName, path):
        self._writeOverlayFile(overlayName, "path.override", "%s\n" % path)

    def existsLayer(self, layerName):
        localLayerDir = os.path.join(self.dir, 'layers', layerName)
        return os.path.exists(localLayerDir)

    def existsOverlay(self, overlayName):
        localOverlayDir = os.path.join(self.dir, 'overlays', overlayName)
        return os.path.exists(localOverlayDir)

    def existsOverlayFile(self, overlayName, fileName):
        localOverlayFile = os.path.join(self.dir, 'overlays', overlayName,
            fileName)
        return os.path.exists(localOverlayFile)

    def wipe(self):
        """
        Wipe out savon's state.
        """
        self.debug('Wiping savon state in %s' % self.dir)
        os.system('rm -rf %s' % self.dir)

