# -*- Mode: Python; test-case-name: savon.test.test_util -*-
# vi:si:et:sw=4:sts=4:ts=4
#
# Savon - sysadmin versioning
# Copyright (C) 2006 Thomas Vander Stichele <thomas at apestaart dot org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.

"""
Utility functions
"""

import os
import shutil
import stat
import pwd
import grp

from savon.configure import configure
from savon.util import command, log

def unique(list):
    """
    Returns the list with every item appearing only once.
    Order is not preserved.
    Similar to list(set(...)), but that is only available in python 2.4
    """
    d = {}
    for item in list:
        if not d.has_key(item):
           d[item] = 1

    return d.keys() 

def mirror(fromPath, toPath):
    """
    Mirror the given fromPath to the given toPath as best as possible.
    Directories are not recursed into.

    This includes:
     - ownership (except for symlinks)
     - permissions (except for symlinks)
     - selinux context
    """
    assert not toPath.endswith(os.path.sep), \
        "Internal error: toPath %s should not end with separator" % toPath
    # toPath's parent needs to exist
    if not os.path.isdir(os.path.dirname(toPath)):
        raise ParentDoesNotExist(os.path.dirname(toPath))

    # first mirror "contents"
    # check for symlink first, because a symlink to a dir satisfies isdir
    if os.path.islink(fromPath):
        fromPathTarget = os.readlink(fromPath)
        if not os.path.lexists(toPath):
            # does not exist yet
            log.debug('mirror', 'Symlinking target %s to %s' % (
                fromPathTarget, toPath))
            os.symlink(fromPathTarget, toPath)
        elif os.path.islink(toPath):
            # exists and is symlink
            toPathTarget = os.readlink(toPath)
            if not fromPathTarget == toPathTarget:
                log.debug('mirror', 'Fixing symlink, target %s to %s' % (
                    fromPathTarget, toPath))
                os.remove(toPath)
                os.symlink(fromPathTarget, toPath)
        else:
            # exists but is not symlink
            os.remove(toPath)
            os.symlink(fromPathTarget, toPath)
    elif os.path.isdir(fromPath):
        # a directory just needs to exist as toPath, there is no content
        # to mirror 
        if os.path.exists(toPath):
            assert os.path.isdir(toPath)
        else:
            os.mkdir(toPath)
    else:
        # source is a file
        if not os.path.exists(toPath):
            log.debug('mirror', 
                'copying %s to non-existent %s' % (fromPath, toPath))
            shutil.copyfile(fromPath, toPath)
        elif not os.path.isfile(toPath):
            log.debug('mirror', 
                'copying %s to non-file %s' % (fromPath, toPath))
            if os.path.isdir(toPath):
                os.rmdir(toPath)
            else:
                os.unlink(toPath)
            shutil.copyfile(fromPath, toPath)
        else:
            fromStat = os.lstat(fromPath)
            toStat = os.lstat(toPath)
            if fromStat.st_size != toStat.st_size \
                or fromStat.st_mtime != toStat.st_mtime:
                log.debug('mirror', 
                    'mtime different, copying %s to pre-existing %s' % (
                        fromPath, toPath))
                os.unlink(toPath)
                shutil.copyfile(fromPath, toPath)

    # os.path.exists returns False for broken symlinks, which for savon
    # purposes is not what we want
    assert os.lstat(toPath)

    fromStat = os.lstat(fromPath)
    toStat = os.lstat(toPath)

    # mirror permissions
    fromPermissions = getPermissions(fromStat)
    toPermissions = getPermissions(toStat)

    if fromPermissions != toPermissions:
        log.log('mirror', "Changing permissions from %04o to %04o on %s" % (
            toPermissions, fromPermissions, toPath))
        os.chmod(toPath, fromPermissions)
    
    # mirror user
    if fromStat.st_uid != toStat.st_uid:
        log.log('mirror', "Changing uid from %s to %s on %s" % (
            toStat.st_uid, fromStat.st_uid, toPath))
        try:
            os.lchown(toPath, fromStat.st_uid, -1)
        except OSError:
            log.error("mirror",
                "Could not change uid on %s from %d to %d" % (
                    toPath, fromStat.st_uid, toStat.st_uid))

    # mirror group
    if fromStat.st_gid != toStat.st_gid:
        log.log('mirror', "Changing gid from %s to %s on %s" % (
            toStat.st_gid, fromStat.st_gid, toPath))
        try:
            os.lchown(toPath, -1, fromStat.st_gid)
        except OSError:
            log.error("mirror",
                "Could not change gid on %s from %d to %d" % (
                    toPath, fromStat.st_gid, toStat.st_gid))

    # mirror selinux context
    if configure.selinux:
        # print 1
        # symlinks have their own context, so make sure we handle that case
        cFrom = configure.selinux.lgetfilecon(fromPath)
        # print 2, toPath
        cTo = configure.selinux.lgetfilecon(toPath)
        # print 3
        if cFrom != cTo:
            log.log("mirror", "changing filecontext from %s to %s on %s" % (
                cTo[1], cFrom[1], toPath))
            try:
                res = configure.selinux.lsetfilecon(toPath, cFrom[1])
                if res == -1:
                    # we can't get at errno from python
                    # setfilecon should be fixed to raise OSError
                    log.warning("mirror",
                        "Could not change SELinux file context "
                        "on '%s' to '%s'" % (toPath, cFrom[1]))
            except TypeError, e:
                log.error("mirror", "SELinux error.  " \
                    "Most likely you are using a version older than 1.30.3.  " \
                    "%r" % e)

def getAbsolutePaths(paths=None):
    """
    Convert the given list of paths to absolute paths.
    If no paths given, return a list with only the current working directory.
    """
    if not paths:
        paths = [os.getcwd(), ]
    return [os.path.abspath(p) for p in paths]

def pathEquals(path1, path2):
    """
    Compare the two given paths.
    Returns True if they're equal in savon terms:
     - same length, content
     - same ownership
     - same permissions
     - same secontext

    filecmp.cmp is not adequate because it compares the whole stat signature,
    including timestamps
    """
    # check uid/gid
    st1 = os.lstat(path1)
    st2 = os.lstat(path2)
    for p in [stat.ST_UID, stat.ST_GID]:
        if st1[p] != st2[p]:
            log.log("util", "user/group different between %s and %s" % (
                path1, path2))
            return False

    # compare permissions
    perms1 = getPermissions(st1)
    perms2 = getPermissions(st2)
    if perms1 != perms2:
        log.log("util", "permissions different between %s (%o) and %s (%o)" % (
            path1, perms1, path2, perms2))
        return False
 
    # compare secontext
    if configure.selinux:
        context = configure.selinux.lgetfilecon(path1)
        secontext1 = context[1]
        context = configure.selinux.lgetfilecon(path2)
        secontext2 = context[1]
        if secontext1 != secontext2:
            log.log("util",
                "secontext different between %s (%s) and %s (%s)" % (
                    path1, secontext1, path2, secontext2))
            return False
        else:
            log.log("util",
                "secontext same between %s (%s) and %s (%s)" % (
                    path1, secontext1, path2, secontext2))

    # directories can have different sizes, due to file system
    if not stat.S_ISDIR(st1.st_mode):
        if st1[stat.ST_SIZE] != st2[stat.ST_SIZE]:
            log.log("util", "size different between %s and %s" % (path1, path2))
            return False

    if os.path.isdir(path1) and os.path.isdir(path2):
        log.log("util", "%s and %s are both directories" % (path1, path2))
        return True

    # if both are symlinks, then we should check their targets
    if os.path.islink(path1) and os.path.islink(path2):
        log.log("util", "%s and %s are both symlinks" % (path1, path2))
        if os.readlink(path1) == os.readlink(path2):
            log.log("util", "%s and %s symlink to same target" % (path1, path2))
            return True
        log.log("util", "%s and %s symlink to different target" % (
            path1, path2))
        return False


    command = "diff --brief %s %s" % (path1, path2)
    value = os.system(command)
    if value == 0:
        log.log("util", "%s and %s have no content diff" % (path1, path2))
        return True

    log.log("util", "%s and %s have content diff" % (path1, path2))
    return False

def diffPathProperties(path1, path2):
    """
    Compare the two given paths for properties we track in savon.

    Returns a dict with the usual keys, and a tuple of (1, 2)
    """
    r = {}
    st1 = os.stat(path1)
    st2 = os.stat(path2)

    uid1 = st1[stat.ST_UID]
    uid2 = st2[stat.ST_UID]
    if uid1 != uid2:
        r['user'] = (pwd.getpwuid(uid1)[0], pwd.getpwuid(uid2)[0])

    gid1 = st1[stat.ST_GID]
    gid2 = st2[stat.ST_GID]
    if gid1 != gid2:
        r['group'] = (grp.getgrgid(gid1)[0], grp.getgrgid(gid2)[0])

    perms1 = getPermissions(st1)
    perms2 = getPermissions(st2)
    if perms1 != perms2:
        r['permissions'] = ("%04o" % perms1, "%04o" % perms2)

    if configure.selinux:
        context = configure.selinux.getfilecon(path1)
        secontext1 = context[1]
        context = configure.selinux.getfilecon(path2)
        secontext2 = context[1]
        if secontext1 != secontext2:
            r['secontext'] = (secontext1, secontext2)

    return r

def getPermissions(statbuf):
    """
    Given the stat buffer, return only the permissions.
    """
    return statbuf.st_mode & (stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)

class LogCommand(command.Command, log.Loggable):
    def __init__(self, parentCommand=None, **kwargs):
        command.Command.__init__(self, parentCommand, **kwargs)
        self.logCategory = self.name

class OverlayConflict(Exception):
    """
    An overlay conflicts.
    args is (overlayName, livePaths).
    """

class ParentDoesNotExist(Exception):
    """
    The parent of the given path does not exist.
    args is (parentPath, )
    """

class ParentNotVersioned(Exception):
    """
    The parent of the given path is not under version control.
    args is (parentPath, )
    """
